from collections.abc import Callable
from typing import Any

from NL2SQLEvaluator.logger import get_logger

logger = get_logger(__name__)

_registry: dict[str, dict[str, Any]] = {}


def register_node(
        package_name: str | None = None,
        fun_name: str | None = None,
) -> Callable[..., Any]:
    def decorator(node: Any) -> Any:
        name = fun_name or getattr(node, "__name__", repr(node))
        module_key = package_name or _package_level_name(node, base_package='NL2SQLEvaluator')

        bucket = _registry.setdefault(module_key, {})
        if name in bucket:
            logger.warning("Class '%s' is already registered in '%s'. Skipping.", name, module_key)
        else:
            bucket[name] = node
        logger.debug(f"Registered node `{name}` under package `{module_key}`")
        return node

    return decorator


def get_node_from_registry(package_name, fun_name: str) -> Any:
    if package_name not in _registry:
        raise KeyError(f"Package '{package_name}' not found in registry.  Available packages: {list(_registry.keys())}")
    bucket = _registry[package_name]
    if fun_name not in bucket:
        raise KeyError(f"Function/Class '{fun_name}' not found in package '{package_name}'.")
    return bucket[fun_name]()


def get_available_functions(package_name) -> list[str]:
    if package_name not in _registry:
        raise KeyError(f"Package '{package_name}' not found in registry. Available packages: {list(_registry.keys())}")
    return list(_registry[package_name].keys())


def _package_level_name(node: Any, *, base_package: str | None = None) -> str:
    """
    Given a class/function, return the top-level subpackage below `base_package`
    (or the immediate parent package if base_package is None).

    Examples:
    - node.__module__ = 'NL2SQLEvaluator.db_executor_nodes.db_executor_protocol'
      base_package='NL2SQLEvaluator'  -> 'db_executor_nodes'
    - node.__module__ = 'pkg.sub.mod' base_package=None -> 'sub'
    """
    mod = getattr(node, "__module__", "") or ""
    # Remove the leaf module name to get the parent package path
    # e.g., 'a.b.c' -> 'a.b'
    parent_pkg = mod.rsplit(".", 1)[0] if "." in mod else mod

    if base_package:
        prefix = base_package + "."
        if parent_pkg.startswith(prefix):
            # Keep only the part after 'base_package.'
            rest = parent_pkg[len(prefix):]  # e.g., 'db_executor_nodes'
            return rest.split(".", 1)[0] or base_package
        # If it doesn't start with base, just fall back to last segment
    # No base_package: return immediate parent segment
    return parent_pkg.split(".")[-1] if parent_pkg else mod



import re
from typing import Protocol, Any, Iterator, Self

from pydantic import BaseModel, model_validator

from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable, SQLCacheProtocol


class ExecutorError(Exception):
    pass


class ExecuteTask(BaseModel):
    db_files: list[str] | str
    queries: list[str]
    params: list[dict] | dict | None = None
    db_ids: list[str] | None = None

    # validate target_sql, inputs seq and db_files all have same length
    @model_validator(mode='after')
    def check_len(self) -> Self:
        if isinstance(self.db_files, str):
            self.db_files = [self.db_files for _ in self.queries]

        if self.params is None:
            self.params = [{} for _ in self.queries]
        elif isinstance(self.params, dict):
            self.params = [self.params for _ in self.queries]

        len_queries = len(self.queries)
        if not len(self.db_files) == len_queries:
            raise ValueError(f"Length of db_files {len(self.db_files)} must match length of queries {len_queries}.")
        if not len(self.params) == len_queries:
            raise ValueError("Length of params must match length of queries.")
        if self.db_ids is not None and len(self.db_ids) == len_queries:
            raise ValueError("Length of db_ids must match length of queries.")

        return self

    def __iter__(self) -> Iterator[Any]:
        return iter(zip(self.queries, self.params, self.db_files))


class DbReaderProtocol(Protocol):
    @staticmethod
    def execute_queries(
            tasks: list[ExecuteTask],
            cache_db: SQLCacheProtocol | None = None,
            cache_db_file: str | None = None,
            *args, **kwargs
    ) -> list[list[OutputTable | ExecutorError]]:
        ...


def get_last_pattern_or_same(generation: str, pattern: str):
    matches = re.findall(pattern, generation, re.DOTALL | re.IGNORECASE)
    if matches:
        return matches[-1].strip()
    else:
        return generation


def extract_sql_or_same(generation: str):
    sql_from_answer_tag = get_last_pattern_or_same(generation, r"<answer>(.*?)</answer>")
    sql_without_quotes = get_last_pattern_or_same(sql_from_answer_tag, r"```sql(.*?)```")
    sql_cleaned = sql_without_quotes.strip().strip("`").strip()
    return sql_cleaned


def execute_queries_in_model_predictions(
        db_executor: DbReaderProtocol,
        db_files: list[str],
        queries: list[list[str]],
        params: list[dict] | None = None,
        sql_cache_protocol: SQLCacheProtocol | None = None,
        cache_db_file: str | None = None,
        *args, **kwargs
) -> list[list[OutputTable | ExecutorError]]:
    assert len(db_files) == len(queries)
    tasks = [
        ExecuteTask(
            db_files=db_files[i],
            queries=[extract_sql_or_same(query) for query in queries[i]],
            params=params[i] if params is not None else None
        )
        for i in range(len(db_files))
    ]

    return db_executor.execute_queries(tasks=tasks, cache_db=sql_cache_protocol, cache_db_file=cache_db_file,
                                       *args, **kwargs)



import dataclasses
import os
import statistics

from NL2SQLEvaluator.config import ScriptArgs, DatasetArgs, PipelineArgs, ModelArgs
from NL2SQLEvaluator.dataset_reader_nodes.data_reader_protocol import read_data_from_file
from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import extract_sql_or_same
from NL2SQLEvaluator.hf_parser import TrlParser
from NL2SQLEvaluator.logger import get_logger
from NL2SQLEvaluator.node_registry import get_node_from_registry
from NL2SQLEvaluator.pipeline import run_pipeline

logger = get_logger(__name__)


@dataclasses.dataclass
class SummaryResults:
    strategy: str
    model_name: str
    dataset_name: str
    metric: str
    value: float
    std: float


def run_evaluation(script_args: ScriptArgs, data_args: DatasetArgs, model_args: ModelArgs, pipeline_args: PipelineArgs):
    logger.info("Args: %s", (script_args, data_args, model_args, pipeline_args))
    # read dataset:
    reader = get_node_from_registry('dataset_reader_nodes', pipeline_args.dataset_reader_node)
    dataset = read_data_from_file(reader, data_args.dataset_path, base_db_path=data_args.relative_db_base_path)
    dataset = dataset[:10]
    if not dataset:
        logger.warning("Empty dataset: %s", data_args.dataset_path)

    num_experiments = script_args.num_of_experiments_to_get_std if model_args.temperature > 0 else 1
    logger.info(f'Running {num_experiments} experiments to calculate standard deviation.')
    db_files, target_sqls, pred_sqls, input_seqs = _prepare_input(dataset, data_args)
    data_input = PipelineInput(
        db_files=db_files * num_experiments,
        target_sql=target_sqls * num_experiments,
        predictions=pred_sqls * num_experiments if data_args.pred_col_name else None,
        input_seq=input_seqs * num_experiments
    )

    data_with_scores = run_pipeline(data_input, script_args, data_args, model_args, pipeline_args)
    df = data_with_scores.to_pandas()

    ex_n = []
    for i in range(num_experiments):
        start = i * len(dataset)
        end = start + len(dataset)
        n_pred = data_with_scores.predictions[start:end]
        n_scores = data_with_scores.scores[start:end]
        df[f'{model_args.model_name_or_path}_{i}'] = n_pred
        df[f'{pipeline_args.evaluator_node}_{i}'] = n_scores
        queries_i = [[extract_sql_or_same(query) for query in query_list] for query_list in n_pred]
        ex_n.append(statistics.mean(n_scores))
        df[f'predicted_SQL_{i}'] = queries_i
        df[f'EX_{i}'] = n_scores

    summary_results = SummaryResults(
        strategy='greedy' if model_args.number_of_completions == 1 else 'majority_voting',
        model_name=model_args.model_name_or_path,
        dataset_name=os.path.basename(data_args.dataset_path),
        metric=pipeline_args.evaluator_node,
        value=statistics.mean(ex_n) * 100,
        std=(statistics.stdev(ex_n) * 100) if len(ex_n) > 1 else 0.0
    )

    logger.warning(summary_results)

    if pipeline_args.saver_node is not None:
        saver = get_node_from_registry('saver_nodes', pipeline_args.saver_node)
        saver.save(script_args.output_dir, df=df,
                   configs=(script_args, data_args, model_args, pipeline_args, summary_results))

    return data_input.scores


def _prepare_input(dataset, data_args):
    db_files = []
    target_sqls = []
    pred_sqls = []
    input_seqs = []
    for item in dataset:
        db_files.append(item['db_file'])

        target_ = item[data_args.target_seq_col_name]
        if isinstance(target_, str):
            target_ = [target_]
        target_sqls.append(target_)

        pred_ = None
        if data_args.pred_col_name:
            pred_ = item[data_args.pred_col_name]
            if isinstance(pred_, str):
                pred_ = [pred_]

        pred_sqls.append(pred_)

        input_seqs.append(item[data_args.input_seq_col_name])

    return db_files, target_sqls, pred_sqls, input_seqs


def main():
    parser = TrlParser(dataclass_types=[ScriptArgs, DatasetArgs, ModelArgs, PipelineArgs])  # or [ScriptArgs]
    script_args, data_args, model_args, pipeline_args = parser.parse_args_and_config()
    run_evaluation(script_args, data_args, model_args, pipeline_args)


if __name__ == "__main__":
    main()



import sqlite3

import pytest

from NL2SQLEvaluator.config import PipelineArgs
from NL2SQLEvaluator.pipeline import PipelineInput, run_pipeline


@pytest.fixture
def db_file(tmp_path):
    """
    Create a temporary sqlite database file under tmp_path, populate it with sample rows,
    and yield the file path as a string.
    """
    db_file = tmp_path / "test.db"
    conn = sqlite3.connect(db_file)
    cur = conn.cursor()
    cur.execute("CREATE TABLE users (id INTEGER PRIMARY KEY, name TEXT, age INTEGER)")
    cur.executemany(
        "INSERT INTO users (name, age) VALUES (?, ?)",
        [
            ("Alice", 30),
            ("Bob", 25),
            ("Carol", 40),
        ],
    )
    conn.commit()
    conn.close()
    yield str(db_file)
    # tmp_path cleanup is handled by pytest; no manual delete required


class TestPipeline:

    def test_no_predictions(self, db_file):
        pipe_input = PipelineInput(
            db_files=[db_file],
            target_sql=[["SELECT * FROM users;"]],
            predictions=[["SELECT * FROM users;"]],
            input_seq=None,
        )

        result = run_pipeline(data_input=pipe_input, pipeline_args=PipelineArgs())
        assert result.scores[0] == 1.0



import time

import requests
from litellm.types.utils import ModelResponse
from vllm import SamplingParams

from NL2SQLEvaluator.dataset_reader_nodes.data_reader_protocol import ChatMessageHF
from NL2SQLEvaluator.logger import get_logger
from NL2SQLEvaluator.node_registry import register_node

logger = get_logger(__name__)


def check_server_availability(base_url, total_timeout: float = 320, retry_interval: float = 60):
    """
    Check server availability with retries on failure. This function is only used with hosted_vllm provider.
    If the server is not up after the total timeout duration, raise a `ConnectionError`.
    """
    url = f"{base_url}/health/"
    start_time = time.time()  # Record the start time

    while True:
        try:
            response = requests.get(url)
        except requests.exceptions.RequestException as exc:
            # Check if the total timeout duration has passed
            elapsed_time = time.time() - start_time
            if elapsed_time >= total_timeout:
                raise ConnectionError(
                    f"The vLLM server can't be reached at {base_url} after {total_timeout} seconds. Make "
                    "sure the server is running by running `vllm serve`."
                ) from exc
        else:
            if response.status_code == 200:
                logger.info(f"Server is up at `{url}`!")
                return None

        # Retry logic: wait before trying again
        logger.info(f"Server is not up yet at `{url}`. Retrying in {retry_interval} seconds...")
        time.sleep(retry_interval)


@register_node()
class LiteLLMPredictor:
    def infer(self,
              model_name,
              multiple_tasks_messages: list[ChatMessageHF],
              sampling_params: SamplingParams,
              *args,
              **kwargs) -> list[list[str]]:
        import litellm
        # litellm._turn_on_debug()
        litellm.request_timeout = 6000  # increase request timeout to 6000 seconds
        litellm_provider = kwargs.get('litellm_provider', 'hosted_vllm')
        logger.info(f"Inferring with LiteLLM. Provider: {litellm_provider}, Model: {model_name}")
        api_base = None
        if litellm_provider == 'hosted_vllm':
            base_url = f"http://{kwargs.get('vllm_server_host', 'localhost')}:{kwargs.get('vllm_server_port', 8000)}"
            api_base = f"{base_url}/v1"
            logger.info(f'Using hosted_vllm with api_base: {api_base}')
            check_server_availability(base_url, total_timeout=500, retry_interval=60)

        # stream=True makes each item in the returned list a streaming iterator
        model_answer = litellm.batch_completion(
            model=f"{litellm_provider}/{model_name}",
            messages=multiple_tasks_messages,
            temperature=sampling_params.temperature,
            max_tokens=sampling_params.max_tokens if 'gpt' not in model_name.lower() else None,
            top_p=sampling_params.top_p,
            n=sampling_params.n,
            max_completion_tokens=sampling_params.max_tokens,
            presence_penalty=sampling_params.presence_penalty,
            frequency_penalty=sampling_params.frequency_penalty,
            api_base=api_base,
            rpm=kwargs.get('rpm', None),
            drop_params=True,
            reasoning_effort='high' if 'gpt' in model_name.lower() else None,
            max_workers=8,
        )

        parsed_responses = self.parse_model_output(model_answer)
        return parsed_responses

    def parse_model_output(
            self, model_answer: list[ModelResponse]
    ) -> list[list[str]]:

        parsed_response: list[list[str]] = []
        for out in model_answer:
            if not isinstance(out, ModelResponse):
                if isinstance(out, BaseException):
                    raise out
                ValueError(f'The output is not of type ModelResponse but type {type(out)}: {out}')

            choices_response = [choice['message']['content'] for choice in out['choices']]

            parsed_response.append(choices_response)

        parsed_response = [
            response
            for response in parsed_response
        ]

        return parsed_response



from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable
from NL2SQLEvaluator.evaluator_nodes.evaluator_protocol import EvaluateTask
from NL2SQLEvaluator.evaluator_nodes.utils import get_majority_voting_values, sort_with_different_types
from NL2SQLEvaluator.logger import get_logger
from NL2SQLEvaluator.node_registry import register_node

logger = get_logger(__name__)


@register_node()
class EXEvaluator:
    def execute_metric(
            self,
            tasks: list[EvaluateTask],
            *args,
            **kwargs
    ) -> list[float]:
        results = [self._ex(task) for task in tasks]
        return results

    def _ex(self, task: EvaluateTask) -> float:
        if len(task.target) == 0 and len(task.predictions) == 0:
            logger.warning('No target and no predictions provided for EX evaluation, returning 1.0')
            return 1.0
        if len(task.target) == 0:
            logger.warning('No target provided for EX evaluation, returning 0.0')

        target = OutputTable(rows=[tuple(sort_with_different_types(row)) for row in task.target[0]])
        pred = get_majority_voting_values(task.predictions, count_cardinality_in_row=True)

        if pred is None:
            return 0.0
        if len(target) == len(pred) == 0:
            return 1.0
        if len(target) != len(pred):
            return 0.0

        return float(set(target) == set(pred))



from typing import Protocol

from pydantic import BaseModel

from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable


class EvaluateTask(BaseModel):
    predictions: list[OutputTable]
    target: list[OutputTable]


class EvaluatorProtocol(Protocol):
    def execute_metric(
            self,
            tasks: list[EvaluateTask],
            *args,
            **kwargs
    ) -> list[float]:
        """
        Each element in predictions and references corresponds to one input example.
        Each prediction or reference is a list of (multiple) output tables.
        """
        ...


def evaluate_target_and_pred(
        evaluator: EvaluatorProtocol,
        multiple_tasks_preds: list[list[OutputTable | ExecutorError]],
        multiple_tasks_tars: list[list[OutputTable | ExecutorError]],
        *args,
        **kwargs
) -> list[float]:
    """
    Evaluate a single pair of predictions and targets using the provided evaluator.
    """
    tasks = [
        EvaluateTask(
            predictions=[pred for pred in preds if isinstance(pred, OutputTable)],
            target=[tar for tar in tars if isinstance(tar, OutputTable)]
        )
        for preds, tars in zip(multiple_tasks_preds, multiple_tasks_tars)
    ]
    scores = evaluator.execute_metric(
        tasks=tasks,
        *args,
        **kwargs
    )
    return scores



import pytest

from NL2SQLEvaluator.db_executor_nodes import SqliteCache, SQLiteDBExecutor
from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecuteTask
from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import NotFoundInCacheError, OutputTable, DataToFetch, \
    DataToCache


@pytest.fixture
def db_file(tmp_path):
    """
    Create a temporary sqlite database file under tmp_path, populate it with sample rows,
    and yield the file path as a string.
    """
    db_file = tmp_path / "test.db"
    yield str(db_file)


@pytest.fixture
def cache_db(db_file) -> SqliteCache:
    return SqliteCache()


def return_rows_in_cache(db_file: str) -> OutputTable:
    executor = SQLiteDBExecutor()
    task = ExecuteTask(
        db_files=[db_file],
        queries=["SELECT * FROM `cache_data`"],
    )
    result = executor.execute_queries(
        tasks=[task]
    )[0][0]
    return result


class TestSqliteCache:
    def test_insert_works(self, cache_db: SqliteCache, db_file):
        query = "SELECT * FROM users WHERE age > 30"
        result = OutputTable(rows=[("Alice", 35), ("Bob", 40)])
        cache_db.set_in_cache(
            db_file,
            data_to_cache=[DataToCache(db_id='test', query=query, result=result)],
        )

        rows = return_rows_in_cache(db_file).rows
        assert len(rows) == 1
        assert rows[0][1] == "test"
        assert rows[0][2] == query
        assert rows[0][3] == result.compress()

    def test_cache_and_retrieve(self, cache_db: SqliteCache, db_file):
        query = "SELECT * FROM users WHERE age > 30"
        result = OutputTable(rows=[("Carol", 40)])
        cache_db.set_in_cache(
            db_file,
            data_to_cache=[DataToCache(db_id='test', query=query, result=result)],
        )

        retrieved_result = cache_db.get_from_cache(db_file, data_to_fetch=[DataToFetch(db_id='test', query=query)])[0]
        assert retrieved_result == result

    def test_multiple_inserts(self, cache_db: SqliteCache, db_file):
        queries_and_results = [
            ("SELECT * FROM users WHERE age > 30", OutputTable(rows=[("Dave", 45)])),
            ("SELECT * FROM users WHERE age < 20", OutputTable(rows=[("Eve", 18)])),
        ]

        data_to_cache = [
            DataToCache(db_id='test', query=query, result=result)
            for query, result in queries_and_results
        ]
        cache_db.set_in_cache(db_file, data_to_cache=data_to_cache)

        for query, expected_result in queries_and_results:
            retrieved_result = cache_db.get_from_cache(db_file, data_to_fetch=[DataToFetch(db_id='test', query=query)])[
                0]
            assert retrieved_result == expected_result

    def test_cache_miss(self, cache_db: SqliteCache, db_file):
        query = "SELECT * FROM users WHERE age > 50"
        retrieved_result = cache_db.get_from_cache(db_file, data_to_fetch=[DataToFetch(db_id='test', query=query)])
        assert isinstance(retrieved_result[0], NotFoundInCacheError)

    def test_multiple_cache_misses(self, cache_db: SqliteCache, db_file):
        queries = [
            "SELECT * FROM users WHERE age > 60",
            "SELECT * FROM users WHERE age < 10",
        ]
        data_to_fetch = [DataToFetch(db_id='test', query=query) for query in queries]
        retrieved_results = cache_db.get_from_cache(db_file, data_to_fetch=data_to_fetch)

        for result in retrieved_results:
            assert isinstance(result, NotFoundInCacheError)

    def test_mixed_cache_hits_and_misses(self, cache_db: SqliteCache, db_file):
        # Insert one query into the cache
        hit_query = "SELECT * FROM users WHERE age > 30"
        hit_result = OutputTable(rows=[("Frank", 55)])
        cache_db.set_in_cache(
            db_file,
            data_to_cache=[DataToCache(db_id='test', query=hit_query, result=hit_result)],
        )

        # Prepare one hit and one miss
        miss_query = "SELECT * FROM users WHERE age < 10"
        data_to_fetch = [
            DataToFetch(db_id='test', query=hit_query),
            DataToFetch(db_id='test', query=miss_query),
        ]

        retrieved_results = cache_db.get_from_cache(db_file, data_to_fetch=data_to_fetch)

        assert retrieved_results[0] == hit_result
        assert isinstance(retrieved_results[1], NotFoundInCacheError)

    def test_duplicate_insert_ignored(self, cache_db: SqliteCache, db_file):
        query = "SELECT duplicate_test"
        result = OutputTable(rows=[("Dup", 1)])
        data = DataToCache(db_id='test', query=query, result=result)

        cache_db.set_in_cache(db_file, data_to_cache=[data])
        cache_db.set_in_cache(db_file, data_to_cache=[data])  # same item again

        rows = return_rows_in_cache(db_file).rows
        assert len(rows) == 1

    def test_empty_result_cached_and_retrieved(self, cache_db: SqliteCache, db_file):
        query = "SELECT empty_result"
        empty_result = OutputTable(rows=[])
        cache_db.set_in_cache(db_file, data_to_cache=[DataToCache(db_id='test', query=query, result=empty_result)])

        retrieved = cache_db.get_from_cache(db_file, data_to_fetch=[DataToFetch(db_id='test', query=query)])[0]
        assert retrieved == empty_result

    def test_large_result(self, cache_db: SqliteCache, db_file):
        query = "SELECT large_result"
        large_rows = [(f"name_{i}", i) for i in range(500)]
        large_result = OutputTable(rows=large_rows)
        cache_db.set_in_cache(db_file, data_to_cache=[DataToCache(db_id='test', query=query, result=large_result)])

        retrieved = cache_db.get_from_cache(db_file, data_to_fetch=[DataToFetch(db_id='test', query=query)])[0]
        assert retrieved == large_result

    def test_retrieve_preserves_request_order(self, cache_db: SqliteCache, db_file):
        q1 = "SELECT order_test_1"
        r1 = OutputTable(rows=[("One", 1)])
        q2 = "SELECT order_test_2"
        r2 = OutputTable(rows=[("Two", 2)])

        cache_db.set_in_cache(db_file, data_to_cache=[
            DataToCache(db_id='test', query=q1, result=r1),
            DataToCache(db_id='test', query=q2, result=r2),
        ])

        # Request in reverse order and ensure results preserve the request order
        fetched = cache_db.get_from_cache(db_file, data_to_fetch=[
            DataToFetch(db_id='test', query=q2),
            DataToFetch(db_id='test', query=q1),
        ])
        assert fetched[0] == r2
        assert fetched[1] == r1

    def test_query_with_special_characters(self, cache_db: SqliteCache, db_file):
        query = "SELECT * FROM users WHERE name = 'O''Reilly' AND note = 'emoji: 🚀\\nnew'"
        result = OutputTable(rows=[("O'Reilly", "emoji: 🚀\nnew")])
        cache_db.set_in_cache(db_file, data_to_cache=[DataToCache(db_id='test', query=query, result=result)])

        retrieved = cache_db.get_from_cache(db_file, data_to_fetch=[DataToFetch(db_id='test', query=query)])[0]
        assert retrieved == result


from dataclasses import field, dataclass

import NL2SQLEvaluator.dataset_reader_nodes  # noqa: F401
import NL2SQLEvaluator.evaluator_nodes  # noqa: F401
import NL2SQLEvaluator.predictor_nodes  # noqa: F401
import NL2SQLEvaluator.saver_nodes  # noqa: F401
from NL2SQLEvaluator.node_registry import get_available_functions


@dataclass
class ScriptArgs:
    output_dir: str | None = field(
        default="./outputs",
        metadata={"help": "Directory where evaluation outputs will be saved"}
    )
    cache_db_file_path: str | None = field(
        default="./sql_cache.sqlite",
        metadata={"help": "Path to the SQLite database file for caching SQL queries"}
    )
    num_of_experiments_to_get_std: int = field(
        default=1,
        metadata={"help": "Number of experiment runs to compute standard deviation over"}
    )
    execution_timeout: int = field(
        default=500,
        metadata={"help": "Timeout in seconds for executing SQL queries against the database"}
    )


@dataclass
class DatasetArgs:
    dataset_path: str | None = field(
        default="data/omnisql/data/train_bird_processed.json",
    )
    relative_db_base_path: str = field(
        default="data/omnisql/data/bird/train/train_databases",
        metadata={"help": "Relative path to the database files directory"}
    )
    pred_col_name: str = field(
        default=None,
        metadata={"help": "Name of the column containing predictions in the dataset"}
    )
    input_seq_col_name: str = field(
        default="input_seq",
        metadata={"help": "Name of the column containing input sequences in the dataset"}
    )
    db_id_col_name: str = field(
        default="db_id",
        metadata={"help": "Name of the column containing database identifiers in the dataset"}
    )
    target_seq_col_name: str = field(
        default='SQL',
        metadata={"help": "Name of the column containing target SQL queries in the dataset"}
    )


@dataclass
class ModelArgs:
    model_name_or_path: str = field(
        default="gpt-3.5-turbo",
        metadata={"help": "Name or path of the language model to use for SQL generation"}
    )
    temperature: float = field(
        default=0.0,
        metadata={"help": "Sampling temperature for the language model"}
    )
    number_of_completions: int = field(
        default=1,
        metadata={"help": "Number of completions to generate per prompt."},
    )
    temperature: float = field(
        default=0.7,
        metadata={"help": "Sampling temperature; lower is more deterministic (≈0.1–0.7)."},
    )
    top_p: float = field(
        default=0.8,
        metadata={"help": "Nucleus sampling: keep smallest set of tokens whose cumulative prob ≥ top_p (0–1]."},
    )
    top_k: int = field(
        default=20,
        metadata={"help": "Top-k sampling: consider only the top_k most likely tokens (0 disables)."},
    )
    repetition_penalty: float = field(
        default=1.1,
        metadata={"help": "Penalty >1.0 discourages repeating the same tokens."},
    )
    frequency_penalty: float = field(
        default=0.0,
        metadata={"help": "Downweights tokens that appear frequently in the output so far (0–2, often 0 for code)."},
    )
    presence_penalty: float = field(
        default=0.0,
        metadata={"help": "Penalizes tokens already present; encourages exploring new tokens (0–2, often 0 for code)."},
    )
    max_new_tokens: int = field(
        default=8000,
        metadata={"help": "Maximum number of new tokens to generate."},
    )
    dtype: str = field(
        default="bfloat16",
        metadata={"help": "Computation dtype (e.g., float16, bfloat16, float32, auto)."},
    )
    tensor_parallel_size: int = field(
        default=1,
        metadata={"help": "Number of GPUs to shard weights across (tensor parallel)."},
    )
    data_parallel_size: int = field(
        default=1,
        metadata={"help": "Number of data-parallel replicas (usually handled by launcher)."},
    )
    vllm_server_host: str = field(
        default="localhost",
        metadata={"help": "vLLM server host (if using vLLM server)."},
    )
    vllm_server_port: int = field(
        default=8000,
        metadata={"help": "vLLM server port (if using vLLM server)."},
    )
    litellm_provider: str | None = field(
        default=None,
        metadata={"help": "Optional provider for LiteLLM. If vllm online serving, use `hosted_vllm`"},
    )
    gpu_memory_utilization: float = field(
        default=0.80,
        metadata={"help": "Target fraction of GPU memory to use (vLLM/Accelerate style runtimes)."},
    )
    max_model_length: int = field(
        default=30_000,
        metadata={"help": "Max total context (prompt + generated) tokens to allow."},
    )


@dataclass
class PipelineArgs:
    dataset_reader_node: str = field(
        default="bird_dev",
        metadata={
            "help": f"Dataset reader node to use for loading the dataset. Available readers:  {get_available_functions('dataset_reader_nodes')}"}
    )

    predictor_node: str | None = field(
        default=None,
        metadata={
            "help": f"Predictor node to use for generating predictions. Available predictors: {get_available_functions('predictor_nodes')} "}
    )
    db_executor_node: str = field(
        default="SQLiteDBExecutor",
        metadata={
            "help": f"Database executor node to use for executing SQL queries. Available executors: {get_available_functions('db_executor_nodes')}"}
    )
    sql_cache_node: str = field(
        default="SqliteCache",
        metadata={
            "help": f"SQL cache node to use for caching generated SQL queries. Available cache nodes: {get_available_functions('db_executor_nodes')}"}
    )
    evaluator_node: str = field(
        default="BirdEXEvaluator",
        metadata={
            "help": f"Evaluator node to use for evaluating predictions. Available evaluators: {get_available_functions('evaluator_nodes')}"}
    )
    saver_node: str | None = field(
        default=None,
        metadata={
            "help": f"Saver node to use for saving evaluation results. Available savers: {get_available_functions('saver_nodes')}"}
    )



from typing import Protocol, TypedDict, Literal


# ---- Types ----
class ChatTurn(TypedDict):
    role: Literal["system", "user", "assistant", "tool", "function"]
    content: str


ChatMessageHF = list[ChatTurn]  # one conversation


class DataInput(TypedDict):
    db_file: str
    target_query: list[str]
    input_seq: ChatMessageHF


class DataReaderProtocol(Protocol):
    @staticmethod
    def read(file_path: str, *args, **kwargs) -> list[DataInput]:
        ...


def enforce_input_type(input_dict: list[dict]):
    for data in input_dict:
        if 'db_file' not in data or 'target_query' not in data or 'input_seq' not in data:
            raise ValueError("Each input data must contain at least 'db_file', 'target_query', and 'input_seq' keys.")


def read_data_from_file(reader: DataReaderProtocol, file_path: str, *args, **kwargs) -> list[DataInput]:
    data = reader.read(file_path, *args, **kwargs)
    enforce_input_type(data)
    return data



import sqlite3

import pytest

from NL2SQLEvaluator.db_executor_nodes import SQLiteDBExecutor
from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable
from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecuteTask


@pytest.fixture
def db_file(tmp_path):
    """
    Create a temporary sqlite database file under tmp_path, populate it with sample rows,
    and yield the file path as a string.
    """
    db_file = tmp_path / "test.db"
    conn = sqlite3.connect(db_file)
    cur = conn.cursor()
    cur.execute("CREATE TABLE users (id INTEGER PRIMARY KEY, name TEXT, age INTEGER)")
    cur.executemany(
        "INSERT INTO users (name, age) VALUES (?, ?)",
        [
            ("Alice", 30),
            ("Bob", 25),
            ("Carol", 40),
        ],
    )
    conn.commit()
    conn.close()
    yield str(db_file)
    # tmp_path cleanup is handled by pytest; no manual delete required


@pytest.fixture
def another_db_file(tmp_path) -> str:
    """
    A second DB for multi-DB scenarios.
    """
    db = tmp_path / "other.db"
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    cur.execute("CREATE TABLE items (id INTEGER PRIMARY KEY, label TEXT, price REAL)")
    cur.executemany(
        "INSERT INTO items (label, price) VALUES (?, ?)",
        [("pen", 1.2), ("notebook", 2.5), ("eraser", 0.8)],
    )
    conn.commit()
    conn.close()
    return str(db)


@pytest.fixture
def executor(db_file) -> SQLiteDBExecutor:
    return SQLiteDBExecutor()


class TestSqliteDbExecutor:
    def test_execute_query_simple(self, executor: SQLiteDBExecutor, db_file):
        task = ExecuteTask(
            db_files=[db_file],
            queries=["SELECT name FROM users WHERE age > 30"],
        )
        result = executor.execute_queries(
            tasks=[task]
        )
        assert result == [[OutputTable(rows=[("Carol",)])]]

    def test_execute_with_params(self, executor: SQLiteDBExecutor, db_file):
        task = ExecuteTask(
            db_files=[db_file],
            queries=["SELECT name FROM users WHERE age >= :min_age ORDER BY age"],
            params=[{"min_age": 30}]
        )
        result = executor.execute_queries(
            tasks=[task]
        )
        assert result == [[OutputTable(rows=[("Alice",), ("Carol",)])]]

    def test_execute_multiple_jobs_and_assembly(self, executor: SQLiteDBExecutor, db_file: str, another_db_file: str):
        # Two jobs, different DBs, multiple queries each
        task_1 = ExecuteTask(
            db_files=[db_file, db_file],
            queries=["SELECT COUNT(*) FROM users",
                     "SELECT name FROM users ORDER BY id LIMIT 1", ],
        )
        task_2 = ExecuteTask(
            db_files=[another_db_file, another_db_file],
            queries=["SELECT COUNT(*) FROM items",
                     "SELECT label FROM items ORDER BY price DESC LIMIT 1", ],
        )
        res = executor.execute_queries(
            tasks=[task_1, task_2]
        )
        # Ensure rectangular shape [jobs][idx]
        assert len(res) == 2 and len(res[0]) == 2 and len(res[1]) == 2
        assert res[0][0] == OutputTable(rows=[(3,)])  # users count
        assert res[0][1] == OutputTable(rows=[("Alice",)])  # first user by id
        assert res[1][0] == OutputTable(rows=[(3,)])  # items count
        assert res[1][1] == OutputTable(rows=[("notebook",)])  # highest price item (2.5)

    def test_empty_queries_returns_empty_list(self, executor: SQLiteDBExecutor):
        task = ExecuteTask(
            db_files=[],
            queries=[],
        )
        res = executor.execute_queries(
            tasks=[task]
        )
        assert res == [[]]

    def test_params_length_mismatch_raises(self, executor: SQLiteDBExecutor, db_file: str):
        with pytest.raises(ValueError):
            executor.execute_queries(
                tasks=[ExecuteTask(
                    db_files=[db_file],
                    queries=["SELECT 1", "SELECT 2"],
                    params=[{"x": 1}])
                ],
                timeout=5,
            )

    def test_invalid_sql_returns_executor_error(self, executor: SQLiteDBExecutor, db_file: str):
        from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
        res = executor.execute_queries(
            tasks=[ExecuteTask(
                db_files=[db_file],
                queries=["SELEC invalid FROM no_table"])]
        )
        assert isinstance(res[0][0], ExecutorError)

    def test_write_query_disallowed_by_default_raises_error(self, executor: SQLiteDBExecutor, db_file: str):
        # Your code sets PRAGMA query_only=ON when allow_write=False (default) so writes should fail.
        res = executor.execute_queries(
            tasks=[ExecuteTask(
                db_files=[db_file],
                queries=["CREATE TABLE should_fail (x INT)"])],
            timeout=5,
            num_cpus=1,
        )
        from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
        assert isinstance(res[0][0], ExecutorError)

    def test_read_then_implicit_rollback_has_no_side_effects(self, executor: SQLiteDBExecutor, db_file: str):
        # Even if a write sneaks in a SELECT transaction, your code rollbacks after SELECT branch.
        # We simulate a harmless read and then verify DB unchanged at the end.
        pre = executor.execute_queries(
            tasks=[ExecuteTask(

                db_files=[db_file],
                queries=["SELECT COUNT(*) FROM users"])],
            timeout=5,
            num_cpus=1,
        )
        assert pre == [[OutputTable(rows=[(3,)])]]

        # Attempt an INSERT using default (should fail due to query_only=ON → ExecutorError)
        insert = executor.execute_queries(
            tasks=[ExecuteTask(
                db_files=[db_file],
                queries=["INSERT INTO users (name, age) VALUES ('Zoe', 22)"])],
            timeout=5,
            num_cpus=1,
        )
        from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
        assert isinstance(insert[0][0], ExecutorError)
        # Verify no change

        post = executor.execute_queries(
            tasks=[ExecuteTask(
                db_files=[db_file],
                queries=["SELECT COUNT(*) FROM users"])],
            timeout=5,
            num_cpus=1,
        )
        assert post == [[OutputTable(rows=[(3,)])]]

    @pytest.mark.parametrize("num_cpus", [1, 2])
    def test_order_is_preserved_with_multiple_tasks(
            self, executor: SQLiteDBExecutor, db_file: str, another_db_file: str, num_cpus: int
    ):
        """
        Your code reconstructs results by (job_id, idx). This ensures deterministic layout
        even though Pool may return out-of-order. Validate shape and indexing.
        """
        res = executor.execute_queries(
            tasks=[
                ExecuteTask(
                    db_files=[db_file, another_db_file],
                    queries=["SELECT name FROM users WHERE age = 25",
                             "SELECT COUNT(*) FROM items"]
                ),
                ExecuteTask(
                    db_files=[db_file, another_db_file],
                    queries=["SELECT COUNT(*) FROM users",
                             "SELECT label FROM items WHERE price < 1.0"]
                )
            ],
            timeout=5,
            num_cpus=num_cpus,
        )
        assert res[0][0] == OutputTable(rows=[("Bob",)])
        assert res[0][1] == OutputTable(rows=[(3,)])
        assert res[1][0] == OutputTable(rows=[(3,)])
        assert res[1][1] == OutputTable(rows=[("eraser",)])

    def test_timeout(self, executor: SQLiteDBExecutor, db_file: str):
        """
        Force func_timeout to raise FunctionTimedOut so we don't rely on slow SQL tricks.
        """
        from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
        res = executor.execute_queries(
            tasks=[ExecuteTask(
                db_files=db_file,
                queries=[
                    """ WITH RECURSIVE cnt(x) AS (SELECT 1
                                                  UNION ALL
                                                  SELECT x + 1
                                                  FROM cnt
                                                  WHERE x < 10e10)
                        SELECT x
                        FROM cnt;
                    """
                ])],
            timeout=10,
            num_cpus=1,
        )
        assert isinstance(res[0][0], ExecutorError)

    def test_multiple_timeout(self, executor: SQLiteDBExecutor, db_file: str):
        """
        Force func_timeout to raise FunctionTimedOut so we don't rely on slow SQL tricks.
        """
        from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
        res = executor.execute_queries(
            tasks=[ExecuteTask(
                db_files=db_file,
                queries=[
                            """ WITH RECURSIVE cnt(x) AS (SELECT 1
                                                          UNION ALL
                                                          SELECT x + 1
                                                          FROM cnt
                                                          WHERE x < 10e10)
                                SELECT x
                                FROM cnt;
                            """
                        ] * 5)],
            timeout=10,
        )
        assert isinstance(res[0][0], ExecutorError)
        assert isinstance(res[0][1], ExecutorError)
        assert isinstance(res[0][2], ExecutorError)
        assert isinstance(res[0][3], ExecutorError)
        assert isinstance(res[0][4], ExecutorError)

    def test_bad_sql_resilient(self, executor: SQLiteDBExecutor, db_file: str):
        from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
        res = executor.execute_queries(
            tasks=[ExecuteTask(
                db_files=db_file,
                queries=[
                    "SELECT * FROM non_existent_table",
                    "SELECT name FROM users WHERE age = 25",
                    "MALFORMED SQL STATEMENT",
                ])],
            timeout=5,
            num_cpus=1,
        )
        assert isinstance(res[0][0], ExecutorError)
        assert res[0][1] == OutputTable(rows=[("Bob",)])
        assert isinstance(res[0][2], ExecutorError)

    def test_get_from_cache(self, executor, db_file, tmp_path):
        from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import DataToCache
        from NL2SQLEvaluator.db_executor_nodes import SqliteCache
        from NL2SQLEvaluator.db_executor_nodes.sqlite_db_executor import _db_id_from_path

        query = "SELECT * FROM users WHERE age > 30"
        result = OutputTable(rows=[("Alice", 35), ("Bob", 40)])

        # first save in cache
        cache_db_file = tmp_path / "cache_test.db"
        cache_db = SqliteCache()

        cache_db.set_in_cache(
            str(cache_db_file),
            data_to_cache=[DataToCache(db_id=_db_id_from_path(db_file), query=query, result=result)]
        )
        # since the actual result is different from cached, we should get the cached one
        task = ExecuteTask(
            db_files=[db_file],
            queries=[query],
        )
        res = executor.execute_queries(
            tasks=[task],
            cache_db=cache_db,
            cache_db_file=str(cache_db_file),
        )

        assert res == [[result]]



from typing import Any

from NL2SQLEvaluator.dataset_reader_nodes.data_reader_protocol import ChatMessageHF
from NL2SQLEvaluator.logger import get_logger
from NL2SQLEvaluator.node_registry import register_node

logger = get_logger(__name__)


@register_node()
class VLLMPredictor:
    def infer(self,
              model_name: str,
              multiple_tasks_messages: list[ChatMessageHF],
              sampling_params: Any,
              *args,
              **kwargs) -> list[list[str]]:
        tp = kwargs.get('tensor_parallel_size', 1)
        dp = kwargs.get('data_parallel_size', 1)
        reasoning_effort = kwargs.get('reasoning_effort', None)
        max_model_len = kwargs.get('max_model_len', 8128)
        logger.info(f'Loading model {model_name} with tp={tp}, dp={dp}, max_model_len={max_model_len}.')
        llm = self._load_model(model_name, tp, dp, max_model_len)
        tokenizer = self._load_tokenizer(model_name)
        chat_prompts = tokenizer.apply_chat_template(multiple_tasks_messages,
                                                     add_generation_prompt=True,
                                                     tokenize=False,
                                                     reasoning_effort=reasoning_effort)
        output = llm.generate(chat_prompts, sampling_params)
        responses = [o.outputs[0].text for o in output]
        responses = [[o] if not isinstance(o, list) else o for o in responses]
        return responses

    def _load_model(self, model_name, tp, dp, max_model_len):
        # https://docs.vllm.ai/en/latest/configuration/optimization.html#performance-tuning-with-chunked-prefill
        from vllm import LLM
        llm = LLM(
            model=model_name,
            dtype="bfloat16",
            tensor_parallel_size=tp,
            max_model_len=max_model_len,
            gpu_memory_utilization=0.92,
            swap_space=42,
            # disable_custom_all_reduce=True,
            trust_remote_code=True,
            data_parallel_size=dp,
            max_num_batched_tokens=16000,
        )

        return llm

    def _load_tokenizer(self, model_name):
        from transformers import AutoTokenizer
        tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
        return tokenizer



import pandas as pd
from pydantic import BaseModel, model_validator, ConfigDict
from typing_extensions import Self

from NL2SQLEvaluator.config import ScriptArgs, DatasetArgs, ModelArgs, PipelineArgs
from NL2SQLEvaluator.dataset_reader_nodes.data_reader_protocol import ChatMessageHF
from NL2SQLEvaluator.db_executor_nodes.cache.cache_protocol import OutputTable
from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import ExecutorError
from NL2SQLEvaluator.evaluator_nodes.evaluator_protocol import evaluate_target_and_pred
from NL2SQLEvaluator.logger import get_logger
from NL2SQLEvaluator.node_registry import get_node_from_registry

logger = get_logger(__name__)


class PipelineTask(BaseModel):
    db_file: str
    target_sql: list[str]
    predictions: list[str] | None = None
    input_seq: ChatMessageHF | None = None
    executed_tar_sqls: list[OutputTable | ExecutorError] | None = None
    executed_pred_sqls: list[OutputTable | ExecutorError] | None = None
    score: float | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # validate target_sql, inputs seq and db_files all have same length
    @model_validator(mode='after')
    def check_len(self) -> Self:
        are_equal = True
        if self.predictions is None and self.input_seq is None:
            raise ValueError("Must provide at least input_seq or predictions.")

        if self.predictions is not None:
            are_equal = are_equal and (len(self.predictions) == len(self.target_sql))
        if self.executed_tar_sqls is not None:
            are_equal = are_equal and (len(self.executed_tar_sqls) == len(self.target_sql))
        if self.executed_pred_sqls is not None:
            are_equal = are_equal and (len(self.executed_pred_sqls) == len(self.target_sql))

        if not are_equal:
            raise ValueError(
                "Length of db_files, target_sql, input_seq, predictions, executed_tar_sqls, and executed_pred_sqls must be equal."
            )
        return self

    def to_pandas(self):
        rows = []
        for i in range(len(self.db_files)):
            row = {
                "db_file": self.db_files[i],
                "input_seq": self.input_seq[i],
                "target_sql": self.target_sql[i],
                "predicted_sql": self.predictions[i] if self.predictions else None,
                "executed_target_sql": self.executed_tar_sqls[i] if self.executed_tar_sqls else None,
                "executed_predicted_sql": self.executed_pred_sqls[i] if self.executed_pred_sqls else None,
                "score": self.scores[i] if self.scores else None,
            }
            rows.append(row)

        return pd.DataFrame(rows)


def run_pipeline(pipeline_tasks=list[PipelineTask],
                 script_args: ScriptArgs = ScriptArgs(),
                 data_args: DatasetArgs = DatasetArgs(),
                 model_args: ModelArgs = ModelArgs(),
                 pipeline_args: PipelineArgs = PipelineArgs()) -> list[PipelineTask]:
    logger.info("Running evaluation with args:", (script_args, data_args, model_args, pipeline_args))
    predictor, cache_db, db_executor, evaluator = _get_pipeline_nodes(pipeline_args)

    data_with_predictions = _get_predictions(pipeline_tasks, predictor, model_args)
    data_with_executed_sqls = _execute_sqls(data_with_predictions, db_executor, cache_db,
                                            script_args.cache_db_file_path,
                                            script_args.execution_timeout)
    return _evaluate(data_with_executed_sqls, evaluator)


def _get_predictions(pipeline_tasks: list[PipelineTask], predictor, model_args: ModelArgs) -> list[PipelineTask]:
    tasks_input_seq = [task.input_seq for task in pipeline_tasks]
    if tasks_input_seq[0] is None:
        return pipeline_tasks

    from NL2SQLEvaluator.predictor_nodes.predictor_protocol import generate_predictions
    predictions = generate_predictions(
        predictor=predictor,
        model_name=model_args.model_name_or_path,
        multiple_tasks_messages=tasks_input_seq,
        model_args=model_args)

    return [task.model_copy(update={"predictions": predictions}) for task in pipeline_tasks]


def _execute_sqls(pipeline_tasks: list[PipelineTask], db_executor, cache_db, cache_db_file, timeout) -> list[
    PipelineTask]:
    from NL2SQLEvaluator.db_executor_nodes.db_executor_protocol import execute_queries_in_model_predictions
    to_be_executed = []
    [to_be_executed.append(task.target_sql) for task in pipeline_tasks]
    [to_be_executed.append(task.predictions) for task in pipeline_tasks]
    db_files = [task.db_file for task in pipeline_tasks] * 2

    executed_queries = execute_queries_in_model_predictions(
        db_executor=db_executor,
        db_files=db_files,
        queries=to_be_executed,
        params=None,
        sql_cache_protocol=cache_db,
        cache_db_file=cache_db_file,
        timeout=timeout,
    )
    executed_tar_sqls = executed_queries[: len(pipeline_tasks)]
    executed_pred_sqls = executed_queries[len(pipeline_tasks):]

    return [
        task.model_copy(update={"executed_tar_sqls": executed_tar_sqls, "executed_pred_sqls": executed_pred_sqls})
        for task in pipeline_tasks
    ]


def _evaluate(pipeline_tasks: list[PipelineTask], evaluator) -> list[PipelineTask]:
    executed_pred = []
    executed_target = []
    for task in pipeline_tasks:
        executed_pred.append(task.executed_pred_sqls)
        executed_target.append(task.executed_tar_sqls)

    evaluation_results = evaluate_target_and_pred(
        evaluator,
        multiple_tasks_preds=executed_pred,
        multiple_tasks_tars=executed_target
    )
    return [task.model_copy(update={"score": evaluation_results}) for task in pipeline_tasks]


def _get_pipeline_nodes(pipeline_args: PipelineArgs):
    predictor = None
    if pipeline_args.predictor_node is not None:
        predictor = get_node_from_registry('predictor_nodes', pipeline_args.predictor_node)
    cache_db = None
    if pipeline_args.sql_cache_node:
        cache_db = get_node_from_registry('db_executor_nodes', pipeline_args.sql_cache_node)

    db_executor = get_node_from_registry('db_executor_nodes', pipeline_args.db_executor_node)
    evaluator = get_node_from_registry('evaluator_nodes', pipeline_args.evaluator_node)
    return predictor, cache_db, db_executor, evaluator
